/********************************************************************  
   This file implements all the necessary Hash Table routines.
*********************************************************************/

#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>

#include "common.h"
#include "globalProcs.h"
#include "globalVars.h"
#include "HashRoutines.h"

/***********************************  
  Declarations of static function 
************************************/

static void freeMemory( struct item *ptr );
static void freeList ( struct ListOfItems *ptr );


/*********************************
   STATIC FUNCTION DEFINITIONS
**********************************/

static void freeList ( struct ListOfItems *ptr )
{
   if (ptr)
   {
      freeList ( ptr -> next );
      free ( ptr );
   }
}


static void freeMemory( struct item *ptr )
{
   if (ptr)
   {
      freeMemory ( ptr -> nextItem );
      freeList ( ptr -> ptrToConItems );
      freeList ( ptr -> ptrToRelationsTo );
      freeList ( ptr -> ptrToRelationsFrom );
      free ( ptr );
   }
}


/***********************************
   EXTERNAL FUNCTION DEFENITIONS
************************************/


void viewHashTable ( struct Data *hash )
{
   int i;
   struct item *p;
   
   for ( i=0; i < HASH_TABLE_SIZE; i++ )
   {
      fprintf ( debugFileNum, "\nHash index : %d\n", i );
      
      p = hash -> HashTable[i];
      while ( p ) 
      {
          fprintf ( debugFileNum, "    %s\n", p -> name );
          p = p -> nextItem;
      }
   }

}

/*******************************************
 calculates index into the HashTable
********************************************/
int hashIndex ( char *token )
{
   int indx = 0;
   int i;
   
   for ( i=0; i < strlen(token); i++ )
      indx = indx + toascii( token[i] );
   
   indx = fmod ( (double) indx , (double) HASH_TABLE_SIZE );
   return indx;
}


/*******************************************
 returns nil if token not found else returns
 pointer to item
********************************************/

struct item * searchHash(struct Data *hash, char *token )
{
   struct item *ptr;
   int indx = hashIndex ( token );
   
   if ( hash -> HashTable [indx] )
   {
      if ( strcmp ( hash -> HashTable[indx]->name , token) == 0 )
      {
         return hash -> HashTable [indx];
      }
      else
      {
         ptr = hash -> HashTable[indx] -> nextItem; 
         while (ptr)
         {
            if ( strcmp ( ptr->name, token) == 0 )
               return ptr;
               
            ptr = ptr -> nextItem;
         
         } /* end while */
         return NULL;
      
      } /* end if */
   }
   else
   {
      return NULL;
   }   /* end if */
}


void addItemToHashTable ( struct Data *hash, struct item *ptrToItem )
{
   struct item *ptr;
   char *token = ptrToItem -> name;
   int indx = hashIndex( token );
   
   if ( ! hash->HashTable[indx] )
      hash -> HashTable[indx] = ptrToItem;
   else
   {
      ptr = hash -> HashTable[indx];
      while(1)
      {
         if ( !ptr->nextItem )
            break;
         
         ptr = ptr -> nextItem;
      }
      
      ptr -> nextItem = ptrToItem;
   }
}


void initHashTable ( struct Data *hash )
{
   int i;
   for ( i=0; i < HASH_TABLE_SIZE; i++ )
      hash -> HashTable[i] = NULL;
}


void cleanUp (struct Data *hash )
{
   int i;
   for ( i=0; i < HASH_TABLE_SIZE; i++ )
   {
      if ( hash -> HashTable[i] )
         freeMemory ( hash -> HashTable[i] );
   }
}


