/*********************************************************************
 This file is responsible for parsing layout files.  The format of 
 layout files is as follows:

entity_name {
      x = 10.5
      y = 56.7
  width = 66.5
 height = 39.9 
}

The order of the attributes ( x,y,height,width ) does not matter.

There can be other attributes as well; the program will simply
ignore them.
*********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "support.h"
#include "common.h"
#include "globalProcs.h"
#include "globalVars.h"
#include "HashRoutines.h"
#include "ParseLayout.h"

/*********************************
   STATIC ABSTRACT DATA TYPES
**********************************/

struct Attr
{
   char name[MAX_TOKEN_SIZE];
   float x;
   float y;
   float width;
   float height;
};

/***********************************  
  Declarations of static function 
************************************/

static void getPositionAttr ( struct Attr * );
static void processFactAttributes ( struct Data * );

/*********************************
   STATIC VARIABLE DEFINITIONS
**********************************/

static FILE *fd;


/*********************************
   STATIC FUNCTION DEFINITIONS
**********************************/

static void getPositionAttr ( attrib )
   struct Attr *attrib;
{
   char token[MAX_TOKEN_SIZE];
   char garbage[MAX_TOKEN_SIZE];
   char attr[MAX_TOKEN_SIZE];
   int attrVal;
   int check = 0;
   int i;

#define XVAL 120        /* ascii value of x */
#define YVAL 121        /* ascii value of y */
#define HEIGHT_VAL 633 /* ascii value of height */
#define WIDTH_VAL 544 /* ascii value of width */

      getToken ( attrib -> name, fd );

      if ( feof(fd) )
	  return;

      if (strcmp(attrib->name, "FACT") == 0) {
	   getToken ( garbage, fd );

	   if (strcmp(garbage, "ATTRIBUTE") == 0) {
		/* rest of line is garbage */
                fgets ( garbage, MAX_TOKEN_SIZE, fd );
		getToken(attrib->name, fd);
                fgets ( garbage, MAX_TOKEN_SIZE, fd );
	   }
	   else if (garbage[0] != '{') {
		printf("Error: Expecting attribute record\n");
		exit(1);
	   }
      }
      else {
          fgets ( garbage, MAX_TOKEN_SIZE, fd ); /* rest of line is garbage */
      }
      
      while (1)
      {
         attrVal = 0;

         fscanf ( fd, "%s", attr );
         
         if ( ! strcmp (attr, "}") )  /* if attr == "}" */
         {
            if ( check == XVAL + YVAL + WIDTH_VAL + HEIGHT_VAL )
               break;
            else
               printf ("\nEntity: %s is missing one or more attributes.\n", attrib -> name );
               exit(1);
         }

         fscanf ( fd, "%s", garbage );  /* equal sign */
         fscanf ( fd, "%s", token );   /* data */
         

         for ( i=0; i < strlen(attr); i++ )
            attrVal += (int) attr[i];

         switch ( attrVal )
         {
            case XVAL : attrib -> x =  atof ( token );
                        check += XVAL;
                        break;
            
            case YVAL : attrib -> y =  atof ( token );
                        check += YVAL;
                        break;

            case WIDTH_VAL : attrib -> width = atof ( token );
                            check += WIDTH_VAL;

                            break;

            case HEIGHT_VAL : attrib -> height = atof ( token );
                             check += HEIGHT_VAL;
                            
                             break;
            
            default :  break;

         } /* end switch */
      } /* end while */
}

static void processFactAttributes ( graph )
   struct Data *graph;
{
   struct item *itemPtr = NULL;
   struct Attr *p = (struct Attr *) xmalloc ( sizeof (struct Attr) );
 
   while ( 1 )
   {
      getPositionAttr( p );

      if ( feof(fd) )
          break;

      if ( debugMode )
      {
        fprintf (debugFileNum, "%-30s %f %f %f %f\n",p->name,p->x,p->y,p->height,p->width);
        fflush (debugFileNum);
      }

      itemPtr = searchHash ( graph, p -> name );

      if ( itemPtr )
      {
         itemPtr -> x = p -> x;
         itemPtr -> y = p -> y;
         itemPtr -> height = p -> height;
         itemPtr -> width = p -> width;
      }
      else
      {
         printf ("\nEntity: %s, exists in layout file but not in the graph file.\n", p -> name );
         printf ("Layout file is not associated with the graph file.\n");
         exit(1);
      }

   }
}


/*********************************************************
  external function defenitions
**********************************************************/

void Control_ParseLayout ( graph, fdin )
   struct Data *graph;
   FILE *fdin;
{
   fd = fdin;
   processFactAttributes ( graph );
}
