/*********************************************************************
  This file contains common definitions, user defined type definitions
  (i.e. struct, enums, typedefs),  #defines, etc. for the 3 programs
  ( lslayout, lsadjust, lsdiff ). 
*********************************************************************/

#ifndef COMMON
#define COMMON

/*** 
- this did not work on Rics machine at waterloo
#define ROUND(x) floor(x+0.5)   
                               
***/

#define ROUND(val) ( (int) ( val + 0.5 ) )

#define FACT_TUPLE_HEADER "FACT TUPLE :"
#define INSTANCE "$INSTANCE"
#define CONTAIN "contain"

#define TRUE 1
#define FALSE 0

#define MAX_LINE_SIZE 4096
#define MAX_TOKEN_SIZE 1024

#define HASH_TABLE_SIZE 997

#define WIDTH_SCALE 0.6
#define HEIGHT_SCALE 0.6
   
#define HEIGHT_TO_WIDTH_RATIO 0.6
#define WIDTH_TO_HEIGHT_RATIO 1.666

typedef int BOOLEAN;

enum layoutModel { nested , cliServ };

enum TypesInModel { client, server, subsystem, subsystemItem };

#if defined(LSADJUST) || defined(LSDIFF)

  enum compressType { none , min, max };

#endif /* LSADJUST || LSDIFF */

#ifdef LSADJUST

  enum statusTypes { newItem, oldItem };

#endif /* LSADJUST */

struct item
{
   char *name;
   struct ListOfItems *ptrToConItems;
   int indxIntoParConItemsArray;
   enum TypesInModel kindOfItemInModel ;
   int numOfItemsContained;
   BOOLEAN isContained;
   int containLevel;
   struct ListOfItems *ptrToRelationsTo;
   int numOfRelationsTo;
   int numOfUniqueRelationsTo;
   struct ListOfItems *ptrToRelationsFrom;
   int numOfRelationsFrom;
   float x;
   float y;
   float width;
   float height;
   struct item *parent;
   struct item *nextItem;

#ifdef LSLAYOUT   
   int desiredPosition;
   int actualPosition;
   int cycleDetNum;
   int level;
#endif

#ifdef LSADJUST
   
   enum statusTypes status;

#endif /* LSADJUST */

#if defined(LSADJUST) || defined(LSDIFF)
   
   float oldX;
   float oldY;
   float oldHeight;
   float oldWidth;

#endif   /* LSADJUST || LSDIFF */

#ifdef LSDIFF

   char *entityClass;
   BOOLEAN collided;
   int numberOfCollisions;

#endif /* LSDIFF */

};

struct ListOfItems
{
   struct item *itemPtr;
   struct ListOfItems *next;
};

struct Data
{
   struct item *HashTable[HASH_TABLE_SIZE];
};

struct SystemPointers
{
   struct item *rootItemPtr;
   struct item *subSystemPtr;
};

#define MAX_CLIENT_OR_SERVER_HEIGHT 50


#ifdef LSLAYOUT

#define MAX_ITEMS_PER_LEVEL 1000

struct realArray
{
   float list [ MAX_ITEMS_PER_LEVEL ]; 
};

struct LevelType
{
   struct item *list [ MAX_ITEMS_PER_LEVEL ];
   int numOfItemsInLevel;
};   

#endif  /* LSLAYOUT */


#if defined(LSADJUST) || defined(LSDIFF)

struct ContainerInfo
{
   float containerWidth;        /* width of container  */
   float containerHeight;       /* height of container */
   
   float maxXitemsInContainer;  /* max x position of the items in container */
   struct item *ptrToMaxXItem;  /* pointer to item that has the max x position */
   
   float containerItemWidth;   
   float containerItemHeight;

#ifdef LSADJUST

   int numOfNewItemsInContainer;   /* number of new entities in container */
   int numOfItemsInContainer;   /* number of entities in container */

#endif /* LSADJUST */

#ifdef LSDIFF
   
   int numOfColItemsInContainer;   /* number of entities that are causing
                                      * a collision in container
                                      */
#endif /* LSDIFF*/
   
   int X;       /*    - used only when compress = CompressType.full
                      - used to compute the number of times 
                        fcn 'shrink' is called for the purpose of 
                        exiting the fcn with a return type of false
                */

   
   float minServerSpace;   /* this is the amount of space in between 
                              the bottom  of the layout window (Wh)
                              (layout window height) and
                              the server entity closest to it, Es,
                              where Es = Es -> y + Es -> height is
                              a maximum compared to all of the other
                              server entities
                              
                              minServerSpace = Wh - Es 
                           */   
   
   float minServerY;       /* min y position of the servers */
   float maxServerY;       /* max y position of the servers */
   float avgServerWidth;
   float avgServerHeight;
   
   float maxClientY;       /* max y position of the clients */
   float avgClientWidth;
   float avgClientHeight;

#ifdef LSADJUST

   int numOfNewServers;
   int numOfServers;
   int numOfNewClients;
   int numOfClients;
   
#endif   /* LSADJUST */

#ifdef LSDIFF

   int numOfCollidedServers;
   int numOfCollidedClients;
   
#endif  /* LSDIFF */
   
   /*  - used only when compress = CompressType.full
       - used to compute the number of times 
         fcn 'shrink' is called for the purpose of 
         exiting the fcn with a return type of false
   */
   int Xclient;
   int Xserver;

};

#endif  /* LSADJUST || LSDIFF */


#ifdef LSDIFF

   #define NEW "new"
   #define OLD "old"
   #define STATIC ""
   #define DUP ".dup"

#endif /* LSDIFF */


#endif  /* COMMON_H */
