/***************************************************************
  Supporting routines.  
  These are routines that should not really be global but
  because they will be used by at least two of the three programs
  (lslayout, lsadjust, lsdiff) were placed together in a common 
  file that can easily be shared by all three programs. 
***************************************************************/


#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "globalVars.h"
#include "globalProcs.h"
#include "HashRoutines.h"
#include "support.h"

/*********************************************
   Static variable declarations
**********************************************/
static char indentSpace[MAX_LINE_SIZE] = "     ";
static char indent[] = "     ";


#define NUM_SHRINK_ITER 100  /* - this is only used when var compress = max 
                                - it is used as an exit condition until
                                  a better exit condition is found
                                - it is used in functions:
                                      1) shrink
                                      2) shrinkClientOrServer
                            */
                            
/*********************************************
   Static function declarations
**********************************************/
static void getTokenContents ( char [], char [], char [], FILE * );
static void outputInfo ( struct item *, FILE * );

#ifdef LSADJUST

   static void identifyNewOld ( struct item *, struct Data * );

#endif LSADJUST

#if defined(LSADJUST) || defined(LSDIFF)

   static void assignType ( struct item *, enum TypesInModel );
   
   static void unshrinkContainedItems ( struct item * );
   static void shrinkContainedItems ( struct item * );
   static void unshrink ( struct item *, int );
   static void unshrinkClientOrServer ( struct item *, struct ContainerInfo *, enum TypesInModel, int, float );

   static BOOLEAN collide ( struct item *, struct item * );
   static BOOLEAN checkForCollisionCliServ ( struct item *, enum TypesInModel );
   static BOOLEAN checkForCollision ( struct item * );
   static BOOLEAN collisionDetectCliServ ( struct item *, struct item *, enum TypesInModel );
   static void restoreAttributeValues ( struct item * );
   static void restoreAttributeValuesCliServ ( struct item *, enum TypesInModel );

#endif /***** LSADJUST || LSDIFF *****/

/*********************************************
   function definitions
**********************************************/

extern void getToken ( char token[MAX_TOKEN_SIZE], FILE *fdin )
{
   char str1[2] = "";
   char str2[2] = "";
   char garbage[MAX_LINE_SIZE];
   BOOLEAN foundToken = FALSE;

   while ( ! feof(fdin) )
   {
      if ( foundToken )
         break;

      fscanf ( fdin ,"%c", str1 );
      while ( isspace( str1[0] ) )
      {
        if ( feof(fdin) )
           return;
        else
           fscanf ( fdin ,"%c", str1 );
      }

      if ( str1[0] == '/' )    /** possibly rest of line is a comment **/
      {
         fscanf ( fdin, "%c", str2 );
         if ( str2[0] == '/' )    /** if true then rest of line is a comment **/
            fgets ( garbage, MAX_LINE_SIZE, fdin );
         else
         {
            getTokenContents ( str1, str2, token, fdin );
            foundToken = TRUE;
         }
      }
      else
      {  
         getTokenContents ( str1, "" , token, fdin );
         foundToken = TRUE;
      }
   }
}

static void getTokenContents ( str1, str2, token, fdin )
   char str1[2];
   char str2[2];
   char token[MAX_TOKEN_SIZE];
   FILE *fdin;
{
     if ( str1[0] == '"' )
     {
        int i = 0;
        
        while (1) 
        {
           if ( feof(fdin) )
              return;
           
           fscanf ( fdin, "%c", str1 );
           if ( str1[0] == '"' )
              break;
           token[i] = str1[0];
           i++;
        }
        
        token[i] = (char) NULL;
     }
     else 
     {
        int i = 0;
        
        token[i] = str1[0];
        i++;
        
        if ( str1[0] == '/' )
        {
           token[i] = str2[0];
           i++;
        }
        
        while (1) 
        {
           if ( feof(fdin) )
              return;
           
           fscanf ( fdin, "%c", str1 );
           if ( isspace( str1[0] ) )
              break;
              
           token[i] = str1[0];
           i++;
        }
        
        token[i] = (char) NULL;
     }
}

extern void initItem ( struct item *ptr )
{
   ptr -> numOfItemsContained = 0;
   ptr -> numOfRelationsTo = 0;
   ptr -> numOfRelationsFrom = 0;
   ptr -> isContained = FALSE;
   ptr -> containLevel = 1;
   
   ptr -> ptrToConItems = NULL;
   ptr -> ptrToRelationsTo = NULL;
   ptr -> ptrToRelationsFrom = NULL;
   ptr -> nextItem = NULL;

   ptr -> kindOfItemInModel = subsystemItem;

#ifdef LSLAYOUT
   ptr -> level = 1;
   ptr -> cycleDetNum = 0;
   ptr -> desiredPosition = 0;
   ptr -> actualPosition = 0;
#endif

#ifdef LSDIFF
   ptr -> entityClass = NULL;
   ptr -> collided = FALSE;
   ptr -> numberOfCollisions = 0;
#endif

}

void createAndSetRoot ( struct Data *hash, struct SystemPointers *graphPtrs )
{
   struct item *root;
   
   root = ( struct item *) xmalloc ( sizeof(struct item) );
   
   root -> name = "$ROOT";

#ifdef LSLAYOUT  
   root -> x = 0.0;
   root -> y = 0.0;
   root -> height = WINDOW_HEIGHT;
   root -> width = WINDOW_WIDTH;
   root -> level = 1;
   root -> cycleDetNum = 0;
#endif   
   
   root -> numOfItemsContained = 0;
   root -> numOfRelationsTo = 0;
   root -> numOfRelationsFrom = 0;
   root -> nextItem = NULL;
   root -> isContained = FALSE;
   root -> containLevel = 0;
   root -> ptrToConItems = NULL;
   root -> ptrToRelationsTo = NULL;
   root -> ptrToRelationsFrom = NULL;
   
   addItemToHashTable ( hash, root );
   
   { 
      struct item *ptr;
      int i;
      
      for ( i=0; i< HASH_TABLE_SIZE; i++ )
      {
         if ( hash -> HashTable[i] )
         {
            ptr = hash -> HashTable[i];
            
            while (1)
            {
               if ( !ptr )
                  break;
               
               if ( ptr != root )
               {
                   if ( !ptr -> isContained )
                   {
                      ptr -> isContained = TRUE;
                      root -> numOfItemsContained ++;
                      addToList ( &root -> ptrToConItems, ptr );
#ifdef LSLAYOUT
                      ptr -> indxIntoParConItemsArray = root -> numOfItemsContained - 1;
#endif
                   }
               }
               ptr = ptr -> nextItem;
            
            } /* end while */
         
         } /* end if */
      
      } /* end for */
      
   } /* end scope */
   
   graphPtrs -> rootItemPtr = root;
   
   if (debugMode)
      fprintf ( debugFileNum, "$ROOT has been created and set.\n");
   
}


/*********************************************************************************
 In the model of the layout that this program is implementing there are clients, 
 servers and a subsystem.  This procedure will find the pointer to the subsystem
 and set the var 'subSystemPtr' in the SystemPointers class.
**********************************************************************************/ 

void setSubSystemPtr ( graph, graphPtrs, inputFile, subsystemName )
   struct Data *graph;
   struct SystemPointers *graphPtrs;
   char *inputFile;
   char *subsystemName;
{
   struct item *subSystemPtr;
   
   if ( subsystemName )
   {
      subSystemPtr = searchHash ( graph, subsystemName );
   }
   else 
   {
      char SSname[MAX_TOKEN_SIZE];
      getSSname (inputFile, SSname);
      subSystemPtr = searchHash ( graph, SSname );
   }

   if ( subSystemPtr )
   {
      graphPtrs -> subSystemPtr = subSystemPtr;

#ifdef LSLAYOUT
      
      subSystemPtr -> desiredPosition = 1;
      subSystemPtr -> actualPosition = 1;

#endif

      subSystemPtr -> kindOfItemInModel = subsystem;

      {
         struct ListOfItems *p = subSystemPtr -> ptrToConItems;
         while(p)
         {
            p -> itemPtr -> kindOfItemInModel = subsystemItem;
            p = p -> next;
         }
      }
   }
   else
   {
      printf ("Subsystem not detected. Cannot continue.\n");
      exit(1);
   }
}


void adjustContainLevelAssignments ( struct ListOfItems *ptr, int currentContainLevel )
{
   struct ListOfItems *p = ptr;
   
   while (p)
   {
      p -> itemPtr -> containLevel = currentContainLevel + 1;
      if ( p -> itemPtr -> ptrToConItems )
         adjustContainLevelAssignments ( p -> itemPtr -> ptrToConItems,  p -> itemPtr -> containLevel );
      
      p = p -> next;
   } 


} /* adjustContainLevelAssignments */


void view ( struct item *ptr , char ind[MAX_LINE_SIZE] )
{
   struct ListOfItems *p = ptr -> ptrToConItems;
   
   while ( p )
   {
      fprintf ( debugFileNum, "%s %s\n", ind, p -> itemPtr -> name );
      fflush(debugFileNum);
      if ( p -> itemPtr -> ptrToConItems )
      {  
         char tmp[MAX_LINE_SIZE];
         strcpy (tmp, indentSpace);
         strcat(tmp, indent); 
         view ( p -> itemPtr,tmp );
      }
      p = p -> next;
   }
}


void viewStructure ( struct SystemPointers *graphPtrs )
{
   struct item *rootItemPtr = graphPtrs -> rootItemPtr;
   
   fprintf ( debugFileNum, "\nStructure of the graph:\n");
   fprintf ( debugFileNum,   "-----------------------\n\n\n");
   fprintf ( debugFileNum, "%s contains :\n" , rootItemPtr -> name );
   fflush(debugFileNum);
   
   view (rootItemPtr,indentSpace);

   fprintf ( debugFileNum,   "---- END\n\n");
   fflush(debugFileNum);
}


void getSSname ( fileName, ssName )
   char *fileName;
   char *ssName;
{
   char pref[MAX_LINE_SIZE];
   char tmp[MAX_LINE_SIZE];

   char *val = (char *) strrchr (fileName, '/');

   if ( val )
      strcpy ( tmp, val+1 );
   else
      strcpy ( tmp, fileName );

   {
     int tmpLen = strlen ( tmp );
     int prefLen;
     if ( strchr (tmp, '.') )
        prefLen = (long) strchr ( tmp, '.') - (long) tmp;
     else
        prefLen = tmpLen;

     strncpy (pref, tmp, prefLen);
     pref[prefLen] = (char) NULL;
   }
   strcpy (ssName,pref);
   strcat (ssName, ".ss" );
}

static void outputInfo ( ptr, fd )
   struct item *ptr;
   FILE *fd;
{
   struct ListOfItems *p = ptr -> ptrToConItems;
   
   while ( p )
   {
      if ( quoteString( p -> itemPtr -> name ) )
         fprintf ( fd, "\"%s\" {\n", p -> itemPtr -> name );
      else
         fprintf ( fd, "%s {\n", p -> itemPtr -> name );
      
      fprintf ( fd, "      x = %f\n", p -> itemPtr -> x );
      fprintf ( fd, "      y = %f\n", p -> itemPtr -> y );
      fprintf ( fd, "  width = %f\n", p -> itemPtr -> width );
      fprintf ( fd, " height = %f\n", p -> itemPtr -> height );
      fprintf ( fd, "}\n\n\n" );
      if ( p -> itemPtr -> ptrToConItems )
         outputInfo ( p -> itemPtr, fd );
      p = p -> next;   
   }
}


void outputLayout ( rootItemPtr, fd )
   struct item *rootItemPtr;
   FILE *fd;
{
   if ( quoteString( rootItemPtr -> name ) )
      fprintf ( fd, "\"%s\" {\n", rootItemPtr -> name );
   else
      fprintf ( fd, "%s {\n", rootItemPtr -> name );
   
   fprintf ( fd, "      x = %f\n", rootItemPtr -> x );
   fprintf ( fd, "      y = %f\n", rootItemPtr -> y );
   fprintf ( fd, "  width = %f\n", rootItemPtr -> width );
   fprintf ( fd, " height = %f\n", rootItemPtr -> height );
   fprintf ( fd, "}\n\n\n" );
   outputInfo ( rootItemPtr, fd );
}

BOOLEAN isInt (char *str)
{
   int i;
   int len = strlen ( str );
   for ( i=0; i<len; i++)
   {
      if ( ! isdigit( str[i] ) )
        return FALSE;
   }
   return TRUE;
} /* end isInt */


void assignParentInfo ( ptr )
   struct item *ptr;
{
   struct ListOfItems *list = ptr -> ptrToConItems;
   
   while ( list )
   {
      list -> itemPtr -> parent = ptr;
      if ( list -> itemPtr -> ptrToConItems )
         assignParentInfo ( list -> itemPtr );
         
      list = list -> next;
   }
}


/**************************************
   function defenitions for LSADUST
****************************************/
#ifdef LSADJUST

static void identifyNewOld ( ptr, oldGraph )
   struct item *ptr;
   struct Data *oldGraph;
{
   struct item *np;
   struct item *op;
   struct ListOfItems *list = ptr -> ptrToConItems;
   
   while (list)
   {
      np = list -> itemPtr;
      op = searchHash ( oldGraph, np -> name );

      if ( op != NULL && 
	   (strcmp (op->parent->name, np->parent->name) == 0) &&
	   np -> kindOfItemInModel == op->kindOfItemInModel)
      {
	 /* Old graph has same id with same parent and same type */
         np -> status = oldItem;
         
         np -> x = op -> x;
         np -> y = op -> y;
         np -> height = op -> height;
         np -> width = op -> width;      

	 if (debugMode) {
             fprintf ( debugFileNum, "%s (old - %f %f %f %f)\n", 
			np->name, np->x, np->y, np->width, np->height);
             fflush(debugFileNum);
         }
      }       
      else
      {
         np -> status = newItem;
	 np -> x      = 0.0;
	 np -> y      = 0.0;
	 np -> height = 3.0;
	 np -> width  = 4.0;

         if (debugMode) {
             fprintf ( debugFileNum, "%s (new)\n", np->name);
             fflush(debugFileNum);
         }
      }
      
      if ( np -> ptrToConItems )
         identifyNewOld ( np, oldGraph );
         
      list = list -> next;
   }
}

void identifyNewAndOldEntities ( oldGraph,oldGraphPtrs,newGraph,newGraphPtrs )
   struct Data *oldGraph;
   struct SystemPointers *oldGraphPtrs;
   struct Data *newGraph;
   struct SystemPointers *newGraphPtrs;
{
   struct item *newRoot = newGraphPtrs -> rootItemPtr;
   struct item *op;
   
   op = searchHash ( oldGraph, newRoot -> name );
   
   if ( op )
   {
      newRoot -> status = oldItem;
      newRoot -> x = op -> x;
      newRoot -> y = op -> y;
      newRoot -> height = op -> height;
      newRoot -> width = op -> width;
   }   
   else
   {
      printf ("\nFatal Error: $ROOT does not exist in old graph.\n");
      exit(1);
   }
  
   if (debugMode) {
	fprintf ( debugFileNum, "Identifying new and old\n");
	fprintf ( debugFileNum, "-----------------------\n\n");
	fflush(debugFileNum);
   }

   identifyNewOld ( newRoot, oldGraph );

   if (debugMode) {
        fprintf ( debugFileNum, "--- END\n\n");
        fflush(debugFileNum);
   }
}

#endif /* LSADJUST */

/**************************************
   function defenitions for LSADUST || LSDIFF
****************************************/

#if  defined (LSADJUST) || defined (LSDIFF)

BOOLEAN isReal ( char *str)
{
   int i;
   int len = strlen (str);
   for ( i=0; i<len; i++ )
   {
      if ( str[i] == '.' )
         { /* don't do anything */ }
      else if ( ! isdigit( str[i] ) )
         return FALSE;
   }   
   return TRUE;
}

static void assignType ( ptr, kind )
   struct item *ptr;
   enum TypesInModel kind;
{
   struct ListOfItems *list = ptr -> ptrToConItems;
   
   while ( list )
   {
      list -> itemPtr -> kindOfItemInModel = kind;
      if ( list -> itemPtr -> ptrToConItems )
         assignType ( list -> itemPtr, kind );
         
      list = list -> next;
   }
}

void identifyClientsAndServers ( rootItemPtr, subSystemPtr )
   struct item *rootItemPtr;
   struct item *subSystemPtr;
{
   struct item *ptr;
   int i;

   for ( i=0; i < rootItemPtr -> numOfItemsContained; i++ )
   {
      ptr = indexList ( rootItemPtr -> ptrToConItems, i );
      
      if ( ptr != subSystemPtr )
      {
         
         if ( ptr -> numOfRelationsFrom == 0 )   /* client */
         {
            ptr -> kindOfItemInModel = client;    
         }
         else if ( ptr -> numOfRelationsTo == 0 )  /* server */
         {
            ptr -> kindOfItemInModel = server;
         }
         else if ( ptr -> numOfRelationsTo != 0 && ptr -> numOfRelationsFrom != 0 )
         {
            if ( ptr -> numOfRelationsTo >= ptr -> numOfRelationsFrom ) /* client */ 
            {
               ptr -> kindOfItemInModel = client;
            }
            else  /* server */
            {
               ptr -> kindOfItemInModel = server;
            } /* end if */
         }
         else
         {
            /* should never get here */ 
         
         } /* end if */
            
      }  /* end if */ 
      
      if ( ptr -> kindOfItemInModel == client )
         assignType ( ptr, client );
      else if ( ptr -> kindOfItemInModel == server )
         assignType ( ptr, server );

   }  /* end for */ 
}




static void unshrinkContainedItems ( containerPtr )
   struct item *containerPtr;
{
  float divisor = 1 - ( SHRINK / 100 );
  struct ListOfItems *list = containerPtr -> ptrToConItems;
  struct item *ptr;
  
  while ( list )
  {
     ptr = list -> itemPtr;

#ifdef LSADJUST
     if ( ptr -> status == oldItem )
#endif

#ifdef LSDIFF
     if ( ! ptr -> collided )
#endif

     {
         ptr -> x = ptr -> x / divisor;
         ptr -> y = ptr -> y / divisor;
         ptr -> height = ptr -> height / divisor;
         ptr -> width = ptr -> width / divisor;
         
         if ( ptr -> ptrToConItems )
            unshrinkContainedItems ( ptr );
     }
     
     list = list -> next;
   }
}

static void shrinkContainedItems ( containerPtr )
   struct item *containerPtr;
{
   struct ListOfItems *list = containerPtr -> ptrToConItems;
   struct item *ptr;
   
   while ( list )
   {
      ptr = list -> itemPtr;

#ifdef LSADJUST
      if ( ptr -> status == oldItem )
#endif

#ifdef LSDIFF
      if ( ! ptr -> collided )
#endif

      {
         ptr -> x = ptr -> x - ( ptr -> x *  SHRINK / 100 );
         ptr -> y = ptr -> y - ( ptr -> y * SHRINK / 100 );
         ptr -> height = ptr -> height - ( ptr -> height * SHRINK / 100 );
         ptr -> width = ptr -> width - ( ptr -> width * SHRINK / 100 );
         
         if ( ptr -> ptrToConItems )
            shrinkContainedItems ( ptr );         
      }
      
      list = list -> next;
   }
}


/*************************************************
    formula for unshrinking :
   
               Xnew
     Xold = --------------------
            1 - ( SHRINK / 100 )
**************************************************/   
   
static void unshrink ( containerPtr, N )
   struct item *containerPtr;
   int N;
{
  float divisor = 1 - ( SHRINK / 100 );
  struct ListOfItems *list = containerPtr -> ptrToConItems;
  struct item *ptr;
  int i = 0;
  
  while ( list && i < N )
  {
     ptr = list -> itemPtr;

#ifdef LSADJUST
     if ( ptr -> status == oldItem )
#endif

#ifdef LSDIFF
     if ( ! ptr -> collided )
#endif

     {
        ptr -> x = ptr -> x / divisor;
        ptr -> y = ptr -> y / divisor;
        ptr -> height = ptr -> height / divisor;
        ptr -> width = ptr -> width / divisor;
         
        if ( ptr -> ptrToConItems )
           unshrinkContainedItems ( ptr );         
     }
     
     i++;
     list = list -> next;
  }
}

/**************************************************************************
    shrink returns true if after shrinking contents the
    dimensions of all the entities are greater then the minimum allowed
    or else it returns false
    formula for % decrease in X is calculated by :
   
       Xnew = Xold -  Xold * ( SHRINK / 100 )
***************************************************************************/   
   
BOOLEAN shrink ( containerPtr, container, MinHeight, MinWidth )
   struct item *containerPtr;
   struct ContainerInfo *container;
   float MinHeight;
   float MinWidth;
{   
    struct ListOfItems *list = containerPtr -> ptrToConItems;
    struct item *ptr;
    int i = 0;
    
    container -> X += 1;    /** used only when compress = CompressType.max **/
    
    while ( list )
    {
       i++;
       ptr = list -> itemPtr;

#ifdef LSADJUST
       if ( ptr -> status == oldItem )
#endif /* LSADJUST */       

#ifdef LSDIFF
       if ( ! ptr -> collided )
#endif /* LSDIFF */
       {   
          if ( compress == max )
             copyAttributeValues ( ptr );
          
          ptr -> x = ptr -> x - ( ptr -> x *  SHRINK / 100 );
          ptr -> y = ptr -> y - ( ptr -> y * SHRINK / 100 );
             
          if ( compress == min )
          {
              ptr -> height = ptr -> height - ( ptr -> height * SHRINK / 100 );
              ptr -> width = ptr -> width - ( ptr -> width * SHRINK / 100 );
                
              if ( ptr -> height < MinHeight && \
                   ptr -> width < MinWidth )
              {
                 unshrink ( containerPtr, i );
                 return FALSE; 
              }
              else  /** entity was shrunk therefore shrink all its contained entities **/
              {
                 if ( ptr -> ptrToConItems )
                    shrinkContainedItems ( ptr );
              }
          }    
          else if ( compress == max )
          {    
              /******
                 - this is the exit condition for the CompressType.max option
                 - when I think of a better one I will implement it but for
                   now this should be sufficient
              *******/   
              if ( container -> X > NUM_SHRINK_ITER )
                 return FALSE; 
              
              if ( ptr -> height - ( ptr -> height * SHRINK / 100 ) > MIN_CONTAINER_ENTITY_HEIGHT && \
                   ptr -> width - ( ptr -> width * SHRINK / 100 ) > MIN_CONTAINER_ENTITY_WIDTH )
              {
                  ptr -> height = ptr -> height - ( ptr -> height * SHRINK / 100 );
                  ptr -> width = ptr -> width - ( ptr -> width * SHRINK / 100 );
                                   
                  if ( ptr -> ptrToConItems )
                     shrinkContainedItems ( ptr );
              }
               
          }  /* end if */
       }  /* end if */
       
       list = list -> next;
    
    } /* end while */
    
    
   /***
    - if compress == max then return FALSE only if a collision is detected.
    - by returning FALSE you stop the SHRINKING/COMPRESSING process
   ***/
 
   if ( compress == max )
   {
      if ( checkForCollision ( containerPtr ) )
         return FALSE;
      else
         return TRUE;
   }

   return TRUE;
}

/*******************************************************************
 formula for unshrinking x, height, width:

            Xnew
  Xold = --------------------
         1 - ( SHRINK / 100 )

 this is the same equation as the one for shrinking
 
 
 formula for unshrinking y if entity kind == server :
 
           Ynew + ( y -> height * (SHRINK/100) ) - windowHeight * (SHRINK/100)
  Yold =   -------------------------------------------------------------------
                         1 - ( SHRINK / 100 )
                         
*******************************************************************/ 

static void unshrinkClientOrServer ( containerPtr, container, kind, N, windowHeight )
   struct item *containerPtr;
   struct ContainerInfo *container;
   enum TypesInModel kind;
   int N;
   float windowHeight;
{
  float divisor = 1 - ( SHRINK / 100 );
  struct ListOfItems *list = containerPtr -> ptrToConItems;
  struct item *ptr;
  int i = 0;
  
  while ( list && i < N )
  {
     ptr = list -> itemPtr;
     if ( kind == ptr -> kindOfItemInModel )
     {
#ifdef LSADJUST     
         if ( ptr -> status == oldItem )
#endif

#ifdef LSDIFF
         if ( ! ptr -> collided )
#endif  
         {
             if ( kind == client )
                ptr -> y = ptr -> y / divisor;
             else if ( kind == server )
                ptr -> y = (  ptr -> y + ( ptr -> height * (SHRINK/100) ) - ( windowHeight * ( SHRINK/100 ) )  ) / divisor;
                /** ptr -> y = ptr -> y - ( container -> minServerSpace * SHRINK / 100 );  **/
             
             ptr -> height = ptr -> height / divisor;
             ptr -> width = ptr -> width / divisor;
         }
     }
     i++;
     list = list -> next;
  }
}


void copyAttributeValues ( ptr )
   struct item *ptr;
{
   ptr -> oldX = ptr -> x;
   ptr -> oldY = ptr -> y;
   ptr -> oldHeight = ptr -> height;
   ptr -> oldWidth = ptr -> width;
}

static void restoreAttributeValuesCliServ ( containerPtr, kind )
   struct item *containerPtr;
   enum TypesInModel kind;
{
   struct ListOfItems *list = containerPtr -> ptrToConItems;
   struct item *ptr;
   
   while ( list )
   {
      ptr = list -> itemPtr;

#ifdef LSADJUST      
      if ( ptr -> kindOfItemInModel == kind && ptr -> status == oldItem )
#endif

#ifdef LSDIFF
      if ( ptr -> kindOfItemInModel == kind && ptr -> collided == FALSE )
#endif
      {
         ptr -> x = ptr -> oldX;
         ptr -> y = ptr -> oldY;
         ptr -> height = ptr -> oldHeight;
         ptr -> width = ptr -> oldWidth;
      }
      list = list -> next;
   }
}

static void restoreAttributeValues ( containerPtr )
   struct item *containerPtr;
{
   struct ListOfItems *list = containerPtr -> ptrToConItems;
   struct item *ptr;
   
   while ( list )
   {
      ptr = list -> itemPtr;

#ifdef LSADJUST
      if ( ptr -> status == oldItem )
#endif

#ifdef LSDIFF
      if ( ! ptr -> collided )
#endif

      {
         ptr -> x = ptr -> oldX;
         ptr -> y = ptr -> oldY;
         ptr -> height = ptr -> oldHeight;
         ptr -> width = ptr -> oldWidth;
      }
      
      list = list -> next;
   }
}


/*******************************************************************
 shrinkClientOrServer returns true if after shrinking contents the
 dimensions of all the entities are greater then the minimum allowed
 or else it returns false
 formula for % decrease on X is calculated by :

    Xnew = Xold -  Xold * ( SHRINK / 100 )
*******************************************************************/ 

BOOLEAN shrinkClientOrServer ( containerPtr, container, kind, windowHeight ) 
   struct item *containerPtr;
   struct ContainerInfo *container;
   enum TypesInModel kind;
   float windowHeight;
{  
  struct ListOfItems *list = containerPtr -> ptrToConItems;
  struct item *ptr;
  int i;
  
  float MIN_HEIGHT;
  float MIN_WIDTH;

  float oldServerHeight;
  
  if ( compress == max )
  {
     if ( kind == client )
        container -> Xclient += 1;
     else if ( kind == server )
        container -> Xserver += 1;
  }
  
  if ( kind == client )
  {
     MIN_HEIGHT = MIN_CLIENT_HEIGHT;
     MIN_WIDTH = MIN_CLIENT_WIDTH;
  }
  else if ( kind == server )
  {
     MIN_HEIGHT = MIN_SERVER_HEIGHT;
     MIN_WIDTH = MIN_SERVER_WIDTH;
  }
  
  while ( list )
  {
     i++;
     ptr = list -> itemPtr;

#ifdef LSADJUST
     if ( ptr -> status == oldItem )
#endif

#ifdef LSDIFF
     if ( ! ptr -> collided )
#endif

     {
         if ( kind == ptr -> kindOfItemInModel )
         {
             if ( compress == max )
                copyAttributeValues ( ptr );
             
             if ( kind == client )
             {
                ptr -> y = ptr -> y - ( ptr -> y * SHRINK / 100 );
             }
             else if ( kind == server )
             {
                /* ptr -> y = ptr -> y + ( container -> minServerSpace * SHRINK / 100 );  */
                
                ptr -> y = ptr -> y + ( (windowHeight - (ptr -> y + ptr -> height)) * SHRINK/100 );
                
                oldServerHeight = ptr -> height;
             }
             
             if ( compress == min )
             {
                 ptr -> height = ptr -> height - ( ptr -> height * SHRINK / 100 );
                 ptr -> width = ptr -> width - ( ptr -> width * SHRINK / 100 );
                 
                 if ( ptr -> height < MIN_HEIGHT && ptr -> width < MIN_WIDTH )
                 {
                    unshrinkClientOrServer (containerPtr, container, kind, i, windowHeight );
                    return FALSE;
                 }
                 else  
                 {     
                 
                    /** also, adjust the y attribute only for servers **/                 
                    
                    if ( kind == server )
                       ptr -> y = ptr -> y + (oldServerHeight - ptr -> height);
                      
                    /** entity was shrunk therefore shrink all its contained entities **/
                    
                    if ( ptr -> ptrToConItems )
                       shrinkContainedItems ( ptr );
                 }
             } 
             else if ( compress == max )
             {    
                 if ( kind == client )
                 {
                    if ( container -> Xclient > NUM_SHRINK_ITER )
                      return FALSE;
                 }   
                 else if ( kind == server )
                 {
                    if ( container -> Xserver > NUM_SHRINK_ITER )
                       return FALSE;
                 }                    
                              
                 if ( ( ptr -> height - ( ptr -> height * SHRINK / 100 ) )  > MIN_HEIGHT &&
                      ( ptr -> width - ( ptr -> width * SHRINK / 100 ) )  > MIN_WIDTH )
                 {
                     ptr -> height = ptr -> height - ( ptr -> height * SHRINK / 100 );
                     ptr -> width = ptr -> width - ( ptr -> width * SHRINK / 100 );
                     
                     if ( kind == server )
                        ptr -> y = ptr -> y + ( oldServerHeight - ptr -> height );
                     
                     if ( ptr -> ptrToConItems )
                        shrinkContainedItems ( ptr );
                 }
             }
         }
     }  /* end if */
     
     list = list -> next;
  
  } /* end while */

  /***
   - if compress == max then return FALSE only if a collision is detected.
   - by returning FALSE you stop the SHRINKING/COMPRESSING process
  ***/
  if ( compress == max )
  {
     if ( checkForCollisionCliServ ( containerPtr, kind ) )
        return FALSE;
     else
        return TRUE;
  }

  return TRUE;    /* this is executed if compress == min */
  
}
   
   
static BOOLEAN checkForCollisionCliServ ( containerPtr, kind )
   struct item *containerPtr;
   enum TypesInModel kind;
{
   struct ListOfItems *list = containerPtr -> ptrToConItems;
   struct item *ptr;
   
   while ( list )
   {
      ptr = list -> itemPtr;

#ifdef LSADJUST
      if ( ptr -> status == oldItem && ptr -> kindOfItemInModel == kind )
#endif

#ifdef LSDIFF
      if ( ptr -> collided == FALSE && ptr -> kindOfItemInModel == kind )
#endif

      {
         if ( collisionDetectCliServ ( ptr, containerPtr, kind ) )
         {
             restoreAttributeValuesCliServ ( containerPtr, kind );
             return TRUE;
         }
      }
      list = list -> next;
   }
   
   return FALSE;
}


static BOOLEAN checkForCollision ( containerPtr )
   struct item *containerPtr;
{
   struct ListOfItems *list = containerPtr -> ptrToConItems;
   struct item *ptr;
   
   while ( list )
   {
      ptr = list -> itemPtr;

#ifdef LSADJUST
      if ( ptr -> status == oldItem )
#endif

#ifdef LSDIFF
      if ( ! ptr -> collided )
#endif

      {
         if ( collisionDetect ( ptr, containerPtr ) )
         {
             restoreAttributeValues ( containerPtr );
             return TRUE;
         }
      }
      list = list -> next;
   }
   
   return FALSE;
}

static BOOLEAN collide ( p, t )
   struct item *p;
   struct item *t;
{
      float pXcenter = p -> x + ( p -> width / 2 );
      float pYcenter = p -> y + ( p -> height / 2 );
      
      if ( ( pXcenter >= t->x ) &&                  
           ( pYcenter >= t->y ) &&                      
           ( pXcenter <= ( t->x + t->width ) )  &&
           ( pYcenter <= ( t->y + t->height ) )  )     /* check if center of box p
                                                            collides with box t */
            return TRUE;
            
      else if ( ( p->x  >=  t->x ) &&
                ( p->y >= t->y ) &&
                ( p->x <= ( t->x + t->width ) ) &&
                ( p->y <= ( t->y + t->height ) ) )     /* check if top left corner
                                                          of box p collides with box t */
            return TRUE;
      
      else if ( ( p->x + p->width ) >= t->x &&
                ( p->y >= t->y ) &&
                ( ( p->x + p->width ) <= ( t->x + t->width ) )  &&
                ( p->y <= ( t->y + t->height ) )  )   /* check top right corner */
          
             return TRUE;
      
      else if ( ( p->x  >=  t->x ) &&
                ( ( p->y + p-> height ) >= t->y ) &&
                ( p->x <= ( t->x + t->width ) ) &&
                ( ( p->y + p->height ) <= ( t->y + t->height ) ) )   /*check bot left corner */
             
             return TRUE;
      
      else if ( ( p->x + p->width ) >= t->x &&
                ( ( p->y + p-> height ) >= t->y ) &&
                ( ( p->x + p->width ) <= ( t->x + t->width ) ) &&
                ( ( p->y + p->height ) <= ( t->y + t->height ) )  )   /* check bot right corner */

             return TRUE;
      
      else
             return FALSE;
      
} 


static BOOLEAN collisionDetectCliServ ( p, ptr, kind )
   struct item *p;
   struct item *ptr;
   enum TypesInModel kind;
{     
   struct ListOfItems *list = ptr -> ptrToConItems;
   struct item *t;
      
   while ( list )
   {
      t = list -> itemPtr;
      
      if ( t -> kindOfItemInModel == kind && t != p )
      {
          if ( collide ( p,t ) || collide ( t,p ) )
              return TRUE;
      }
      list = list -> next;
   }
 
   return FALSE;
}

/**
  Returns TRUE if entity 'p' collides with any entities
  contained by 'ptr'.   
**/

BOOLEAN collisionDetect ( p, ptr )
   struct item *p;
   struct item *ptr;
{     
   struct ListOfItems *list = ptr -> ptrToConItems;
   struct item *t;
      
   while ( list )
   {
      t = list -> itemPtr;
      
      if ( t != p )
      {
          if ( collide ( p,t ) || collide ( t,p ) )
              return TRUE;
      }
      list = list -> next;
   }
 
   return FALSE;
}


#endif  /***** LSADJUST || LSDIFF *****/
