/************************************************************
   This file is responsible for laying out the clients
   and the servers of a subsystem.  Execution gets here only
   if the layout model selected is the 'client server'
   layout model.
*************************************************************/

#include <stdio.h>
#include <math.h>

#include "./../common/common.h"
#include "./../common/globalProcs.h"
#include "./../common/globalVars.h"
#include "CliServLayout.h"

/***********************************
  Declarations of static function
************************************/

static void calcDesiredPositions ( struct item *, struct item *, int *, int * );
static void calcActualPositionsClients ( struct item *, int );
static void calcActualPositionsServers ( struct item *, int );
static void printClientsAndServers ( struct item *, struct item * );


/*********************************************************
  static function definitions
**********************************************************/


static void calcDesiredPositions ( rootItemPtr, subSystemPtr, numOfClients, numOfServers )
   struct item *rootItemPtr;
   struct item *subSystemPtr;
   int *numOfClients;
   int *numOfServers;
{
   struct item *ptr;
   struct item *itemPtr;
   int sumOfPos;
   int cnt;
   int i,j;

   for ( i=0; i < rootItemPtr -> numOfItemsContained; i++ )
   {
      sumOfPos = 0;
      cnt = 0;
      ptr = indexList ( rootItemPtr -> ptrToConItems, i );
      
      if ( ptr != subSystemPtr )
      {
         
         if ( ptr -> numOfRelationsFrom == 0 )   /* client */
         {
            (*numOfClients)++;
            ptr -> kindOfItemInModel = client;
            
            for ( j=0; j < ptr -> numOfRelationsTo; j++ )
            {
               itemPtr = indexList ( ptr -> ptrToRelationsTo, j );
               if ( itemPtr -> parent == subSystemPtr )
               {
                  sumOfPos += itemPtr -> desiredPosition;
                  cnt++;
               }
            }  /* end for */
         }
         else if ( ptr -> numOfRelationsTo == 0 )  /* server */
         {
            (*numOfServers)++;
            ptr -> kindOfItemInModel = server;
            
            for ( j=0; j < ptr -> numOfRelationsFrom; j++ )
            {
               itemPtr = indexList ( ptr -> ptrToRelationsFrom, j );
               if ( itemPtr -> parent == subSystemPtr )
               {
                  sumOfPos += itemPtr -> desiredPosition;
                  cnt++;
               }
            }  /* end for */
         }
         else if ( ptr -> numOfRelationsTo != 0 && ptr -> numOfRelationsFrom != 0 )
         {
            if ( ptr -> numOfRelationsTo >= ptr -> numOfRelationsFrom ) /* client */ 
            {
               (*numOfClients)++;
               ptr -> kindOfItemInModel = client;
            
               for ( j=0; j < ptr -> numOfRelationsTo; j++ )
               {
                  itemPtr = indexList ( ptr -> ptrToRelationsTo, j );
                  if ( itemPtr -> parent == subSystemPtr )
                  {
                     sumOfPos += itemPtr -> desiredPosition;
                     cnt++;
                  }
               }  /* end for */
            }
            else  /* server */
            {
               (*numOfServers)++;
               ptr -> kindOfItemInModel = server;
            
               for ( j=0; j < ptr -> numOfRelationsFrom; j++ )
               {
                  itemPtr = indexList ( ptr -> ptrToRelationsFrom, j );
                  if ( itemPtr -> parent == subSystemPtr )
                  {
                     sumOfPos += itemPtr -> desiredPosition;
                     cnt++;
                  }
               }  /* end for */
            } /* end if */
         }
         else
         {
            /* should never get here */ 
         
         } /* end if */
            
      }  /* end if */ 

      if ( cnt != 0 )
         ptr -> desiredPosition =  ROUND(sumOfPos / cnt);
      else
         ptr -> desiredPosition = 1;

   }  /* end for */ 
}


static void calcActualPositionsClients ( rootItemPtr, numOfClients )
   struct item *rootItemPtr;
   int numOfClients;
{
   struct item *ptr;
   int desPos;
   int actPos;
   int cliDist;
   int i;

   struct item *cliPosition[numOfClients];

   for ( i=0; i < numOfClients; i++ )
      cliPosition[i] = NULL;

   for ( i=0; i < rootItemPtr -> numOfItemsContained; i++ )
   {
      ptr = indexList ( rootItemPtr -> ptrToConItems, i );

      if ( ptr -> kindOfItemInModel == client )
      {
         desPos = ptr -> desiredPosition - 1;
         if ( desPos >= numOfClients )
            desPos = numOfClients - 1;
      }

      actPos = desPos;

      if ( ptr -> kindOfItemInModel == client )
      {
         cliDist = 1;
         
         while ( cliPosition[actPos] )
         {
            if ( desPos + cliDist < numOfClients && cliPosition[desPos+cliDist] == NULL )
               actPos = desPos + cliDist;
            else if ( desPos - cliDist >= 0 && cliPosition[desPos-cliDist] == NULL )
               actPos = desPos - cliDist;

            cliDist++;      

         }  /* end while */

         ptr -> actualPosition = actPos + 1;
         cliPosition[actPos] = ptr;

      } /* end if */

   } /* end for */  
}


static void calcActualPositionsServers ( rootItemPtr, numOfServers )
   struct item *rootItemPtr;
   int numOfServers;
{
   struct item *ptr;
   int desPos;
   int actPos;
   int servDist;
   int i;

   struct item *servPosition[numOfServers];

   for ( i=0; i < numOfServers; i++ )
      servPosition[i] = NULL;

   for ( i=0; i < rootItemPtr -> numOfItemsContained; i++ )
   {
      ptr = indexList ( rootItemPtr -> ptrToConItems, i );

      if ( ptr -> kindOfItemInModel == server )
      {
         desPos = ptr -> desiredPosition - 1;
         if ( desPos >= numOfServers )
            desPos = numOfServers - 1;
      }

      actPos = desPos;

      if ( ptr -> kindOfItemInModel == server )
      {
         servDist = 1;
         
         while ( servPosition[actPos] )
         {
            if ( desPos + servDist < numOfServers && servPosition[desPos+servDist] == NULL )
               actPos = desPos + servDist;
            else if ( desPos - servDist >= 0 && servPosition[desPos-servDist] == NULL )
               actPos = desPos - servDist;

            servDist++;      

         }  /* end while */

         ptr -> actualPosition = actPos + 1;
         servPosition[actPos] = ptr;

      } /* end if */

   } /* end for */  
}


static void printClientsAndServers ( rootItemPtr, subSystemPtr )
   struct item *rootItemPtr;
   struct item *subSystemPtr;
{
   struct ListOfItems *ptr = rootItemPtr -> ptrToConItems;
   
   fprintf ( debugFileNum, "\n\n Listing Clients Servers and Subsystem Entities.\n" );
   fprintf ( debugFileNum,      "-----------------------------------------------\n\n");

   fprintf ( debugFileNum, "\n\nClients\n" );
   fprintf ( debugFileNum, "-------\n" );

   while ( ptr )
   {
      if ( ptr -> itemPtr -> kindOfItemInModel == client )
         fprintf ( debugFileNum, "%s\n", ptr -> itemPtr -> name );

      ptr = ptr -> next;
   }

   ptr = rootItemPtr -> ptrToConItems;
   fprintf ( debugFileNum, "\n\nServers\n" );
   fprintf ( debugFileNum, "--------\n" );

   while ( ptr )
   {
      if ( ptr -> itemPtr -> kindOfItemInModel == server )
         fprintf ( debugFileNum, "%s\n", ptr -> itemPtr -> name );

      ptr = ptr -> next;
   }

   ptr = subSystemPtr -> ptrToConItems;
   fprintf ( debugFileNum, "\n\nsubsystem entities\n" );
   fprintf ( debugFileNum, "------------------\n" );

   while ( ptr )
   {
      fprintf ( debugFileNum, "%s\n", ptr -> itemPtr -> name );
      ptr = ptr -> next;
   }

}

/*********************************************************
  external function defenitions
**********************************************************/

void Control_CliServLayout (rootItemPtr, subSystemPtr, numOfClients, numOfServers )
   struct item *rootItemPtr;
   struct item *subSystemPtr;
   int *numOfClients;
   int *numOfServers;
{
   calcDesiredPositions ( rootItemPtr, subSystemPtr, numOfClients, numOfServers ); 

   if ( *numOfClients > 0 )
      calcActualPositionsClients ( rootItemPtr, *numOfClients );

   if ( *numOfServers > 0 )
      calcActualPositionsServers ( rootItemPtr, *numOfServers );

   if ( debugMode )
      printClientsAndServers( rootItemPtr, subSystemPtr );
}
