/******************************************************
  This file contains and calls the
  routines that do the actual layout.
******************************************************/

#include <stdio.h>
#include <malloc.h>
#include <assert.h>

#include "./../common/common.h"
#include "./../common/globalProcs.h"
#include "./../common/globalVars.h"
#include "SugiyamaLayout.h"
#include "LevelLayout.h"
#include "CliServLayout.h"
#include "PosDet.h"
#include "DetermineLayout.h"

/***********************************  
  Declarations of static function 
************************************/

static int isThereARelation ( struct item *ptr, int i, int j);
static void buildAdjacencyMatrix ( int **adjacencyMatrix, struct item *containerPtr );
static void printMatrix ( int **aM , struct item *ptr );
static void calcUniqueRelationsTo ( int **aM, struct item *ptr );
static void sortEntitiesByUniqueRelationsTo ( struct item *ptr );
static void updateGlobCycleDetNumOfStackItems();
static void DFS ( struct item *ptr, int **aM, int itemNumber, int level, int *numOfLevels );
static void determineLevelsInContainer ( struct item *ptr, int **aM, int *numOFLevels );
static void printLevels ( struct item *ptr );
static void printLevels ( struct item *ptr );
static void initLevelList ( struct LevelType LevelList[], int numOfLevels );
static void buildLevelStructure ( struct LevelType LevelList[], struct item *ptr );
static void printLevelStructure ( struct LevelType LevelList[], int numOfLevels );
static int maxNumItemsInALevel ( struct LevelType LevelList[] , int numOfLevels );
static void copyMatrix ( int **dest, int **source, int dimension );
static void visitEveryItemThatContainsStuff ( struct item *, struct item * );
static void layoutContainedEntities ( struct item * );


/********************************************
  static variable defenitions
********************************************/

static BOOLEAN cycleDetected = FALSE;

static struct item *curRoot;
static int globCycleDetNum = 0;

#define STACK_SIZE 200

static struct item *dfsStack[STACK_SIZE];
static int dfsTop = 0;



/**********************************************************
  static function defenitions
***********************************************************/



/**********************************************************
  Returns 1 if a relation exists from the i'th to the j'th
  item in the list 'ptr -> ptrToConItems'.
  Else  function returns 0.
***********************************************************/

static int isThereARelation ( struct item *ptr, int i, int j)
{

   struct item *ptrToConI = indexList ( ptr -> ptrToConItems, i );
   struct item *ptrToConJ = indexList ( ptr -> ptrToConItems, j );
   struct item *ptrToRelIK;
   int k;
   
   for ( k=0; k < ptrToConI -> numOfRelationsTo; k++ )
   {
      ptrToRelIK = indexList ( ptrToConI -> ptrToRelationsTo, k );
      if ( ! strcmp ( ptrToRelIK -> name, ptrToConJ -> name ) )
         return 1;
   }
   return 0;
}   


/******************************************************
  Adjacency matrix is built for entities contained by
  containerPtr.
*******************************************************/

static void buildAdjacencyMatrix ( int **adjacencyMatrix, struct item *containerPtr )
{
   int i;
   int j;
   int numOfItemsCon = containerPtr -> numOfItemsContained;
   
   for ( i=0; i < numOfItemsCon; i++ )
   {
      for ( j=0; j < numOfItemsCon; j++ )
      {
          (int) *(adjacencyMatrix + j + i * numOfItemsCon) = isThereARelation ( containerPtr, i, j );      
      }
   }   
}

/**************************************************************
  Prints adjacency matrix 'aM' which represents the entities
  contained by 'ptr'.

  This is used for debugging.
**************************************************************/

static void printMatrix ( int **aM , struct item *ptr )
{
   fprintf (debugFileNum, "\n\nADJACENCY MATRIX OF ENTITIES Of ITEMS IN CONTAINER : %s \n\n", ptr -> name );
   
   {
      int numOfItemsCon = ptr -> numOfItemsContained;
      struct ListOfItems *p = ptr -> ptrToConItems;
      int i;
      int j;
      
      for ( i=0; i < numOfItemsCon; i++ )
      {
         for ( j=0; j < numOfItemsCon; j++ )
         {
            if ( j == 0 )
               fprintf ( debugFileNum, "%3d %3d %-20s",p -> itemPtr -> numOfUniqueRelationsTo, p-> itemPtr -> indxIntoParConItemsArray, p -> itemPtr -> name );
            
            fprintf ( debugFileNum, " %d ", (int) *(aM + j + i * numOfItemsCon) );  
         }
         fprintf ( debugFileNum, "\n");
         p = p -> next;
      }
   }
}


/**************************************************************
   Calculates 'unique relations to' for each entity contained by
  'ptr'.  This is done by counting the number of 1's in each
   row of the adjacency matrix 'aM'. 

   The number of 'unique relations to' is the number of relations
   whose subjects are equal but whose objects are different. 
   
   subject - 2'nd field in tuple
   object - 3'rd field in tuple
   
   For example if there are 2 relations:
   
   useproc file1 file2
   usevar  file1 file2

   The total 'relations to' for entity 'file1' is 2 however
   the number of 'unique relations to' for file1 is 1.
**************************************************************/

static void calcUniqueRelationsTo ( int **aM, struct item *ptr )
{
   int uniqueRels = 0;
   int i;
   int j;
   struct item *p;
   int numOfItemsCon = ptr -> numOfItemsContained;
   
   for ( i=0; i < numOfItemsCon; i++ )
   {
      for ( j=0; j < numOfItemsCon; j++ )
      {
         if ( *(aM + j + i * numOfItemsCon ) )
            uniqueRels++;
      }
      
      p = indexList ( ptr -> ptrToConItems, i );
      p -> numOfUniqueRelationsTo = uniqueRels;
      uniqueRels = 0;
   }
}



/**************************************************************
  Sorts entities in list 'ptr -> ptrToConItems' by the field
  'numOfUniqueRelationsTo'.

  This is done to increase the number of 1's in the right
  side triangle of the adjacency matrix.

  If a line is drawn from the top left corner of a matrix to 
  the bottom right corner of a matrix then two triangles are
  created. The right side triangle is the right most triangle. 
**************************************************************/

static void sortEntitiesByUniqueRelationsTo ( struct item *ptr )
{
   int N = ptr -> numOfItemsContained - 1;
   
   if ( N <= 1 )
      return;
      
   {  
      /** insert dummy node when bubble sorting linked list **/      
      struct ListOfItems *dummy;
      dummy = (struct ListOfItems *) xmalloc ( sizeof(struct ListOfItems) );
      dummy -> next = ptr -> ptrToConItems;
      ptr -> ptrToConItems = dummy;
   } 
   
   {
      int i;
      int j;
      
      struct ListOfItems *t1;
      struct ListOfItems *t2;
   
      struct ListOfItems *p1 = ptr -> ptrToConItems;
      struct ListOfItems *p2 = ptr -> ptrToConItems -> next;
      struct ListOfItems *p3 = ptr -> ptrToConItems -> next -> next;
      
      int tmpIndx;
      
      for ( i=0; i < N - 1; i++ )
      {
         p1 = ptr -> ptrToConItems;
         p2 = ptr -> ptrToConItems -> next;
         p3 = ptr -> ptrToConItems -> next -> next;
         
         for ( j=0; j < N - i; j++ )
         {
            if ( p2 -> itemPtr -> numOfUniqueRelationsTo < p3 -> itemPtr -> numOfUniqueRelationsTo )
            {
               t1 = p3 -> next;
               p1 -> next = p3;
               p3 -> next = p2;
               p2 -> next = t1;
               t2 = p3;
               p3 = p2;
               p2 = t2;
               tmpIndx = p2 -> itemPtr -> indxIntoParConItemsArray;
               p2 -> itemPtr -> indxIntoParConItemsArray = p3 -> itemPtr -> indxIntoParConItemsArray;
               p3 -> itemPtr -> indxIntoParConItemsArray = tmpIndx;
            }
            p1 = p1 -> next;
            p2 = p2 -> next;
            p3 = p3 -> next;
        
        } /* end for */
        
      } /* end for */
   } /* end scope */
        
   {  
      /** free the dummy node in linked list **/      
      struct ListOfItems *tmp;
      tmp = ptr -> ptrToConItems;
      ptr -> ptrToConItems = ptr -> ptrToConItems -> next;
      free (tmp);   
   }

}

/******************************************************
   All the entities on tha dfsStack will get there
   cycle detection number updated so that a cycle
   condition won't incorrectly be detected.
******************************************************/ 

static void updateGlobCycleDetNumOfStackItems()
{
   if ( dfsTop - 1 >= 0 )
   {
      int i;
      for ( i=0; i <= dfsTop - 1; i++ )
         dfsStack[i] -> cycleDetNum = globCycleDetNum;
   }

}

/************************************************************************
This is a Depth First Search Algorithm that assigns levels to the 
entities.  This algorithm also detects cycles.  When a cycle is
detected, the edge that caused the cycle is removed.
************************************************************************/

static void DFS ( struct item *ptr, int **aM, int itemNumber, int level, int *numOfLevels )
{
   struct item *PitemNumber;
   struct item *Pi;
   int numOfItemsCon = ptr -> numOfItemsContained;
   int i;
   
   for ( i=0; i < numOfItemsCon; i++ )
   {
       PitemNumber = indexList ( ptr -> ptrToConItems, itemNumber );
       Pi = indexList ( ptr -> ptrToConItems, i );
       
       if ( curRoot == PitemNumber )
       {
          globCycleDetNum ++;
          PitemNumber -> cycleDetNum = globCycleDetNum;
       }
       
       if ( i != itemNumber )
       {
          if ( (int) *(aM + i + itemNumber * numOfItemsCon) == 1 )
          {
             if ( Pi -> cycleDetNum != globCycleDetNum )
             {
                if ( Pi -> level < level + 1 )
                {
                   Pi -> level = level + 1;
                   
                   if ( *numOfLevels < level + 1 )
                      *numOfLevels = level + 1;
                      
                   Pi -> cycleDetNum = globCycleDetNum;
                   
                   if ( debugMode )
                   {
                      fprintf ( debugFileNum, "DFS : %d   %d \n", itemNumber, i );
                      fprintf ( debugFileNum, "%-30s Level : %d \n", Pi -> name, Pi -> level );
                   }
                   
                   dfsStack [dfsTop] = Pi;
                   dfsTop++;
                   DFS ( ptr, aM, i, level + 1, numOfLevels );
                }
             }
             else
             {
                *(aM + i + itemNumber * numOfItemsCon) = 0;
                if ( debugMode ) 
                   fprintf ( debugFileNum, "Removeing edge :  row : %5d  column : %5d \n", itemNumber, i ); 
                
                cycleDetected = TRUE;
             }
          } /* end if */
       }  /* end if */
   } /* end for */
   
   dfsTop--;
   dfsStack[dfsTop] = NULL;
   globCycleDetNum++;
   updateGlobCycleDetNumOfStackItems();
}


/********************************************************************
   This function determines the number of levels in  container 'ptr'.

   For example if there are two entities A and B, and A calls B then
   A would get assigned to level 1 and B would be assigned to 
   level 2.
**********************************************************************/

static void determineLevelsInContainer ( struct item *ptr, int **aM, int *numOfLevels )
{
   if ( debugMode )
   {
      fprintf ( debugFileNum, "\n\nDetermining the level of the entities in container : %s \n", ptr -> name );
      fprintf ( debugFileNum, "\nTracing through levels of recursion to see how levels are assigned.\n");
   }
   
   {
      int i;
      
      for ( i=0; i < ptr -> numOfItemsContained; i ++ )
      {
         curRoot = indexList ( ptr -> ptrToConItems, i );
         if ( debugMode )
            fprintf ( debugFileNum, "determineLevelsInContainer : %s  Level : %d \n",  curRoot -> name, curRoot -> level );
         
         dfsStack [dfsTop] = curRoot;
         dfsTop++;
         DFS ( ptr, aM, i, curRoot -> level, numOfLevels );
      }
   }  /* end scope */

   if ( cycleDetected )
      if ( debugMode )
         fprintf ( debugFileNum, "WARNING: Cycles were detected in the graph. The layout might not appear very nice.\n");
}



/********************************************************************
  Prints level information to debug file for entities contained
  by 'ptr'.
**********************************************************************/

static void printLevels ( struct item *ptr )
{
   int i;
   struct item *p;
   fprintf ( debugFileNum, "\n\n Level assigned to each entity within container: %s\n", ptr -> name );

   for ( i=0; i < ptr -> numOfItemsContained; i++ )
   {
      p = indexList ( ptr -> ptrToConItems, i );
      fprintf ( debugFileNum, "%-30s %5d\n", p -> name, p -> level );
   }
}


/********************************************************************
   Initialize 'LevelList' array.
**********************************************************************/

static void initLevelList ( struct LevelType LevelList[], int numOfLevels )
{
   int i;
   
   for ( i=0; i < numOfLevels; i++ )
      LevelList[i].numOfItemsInLevel = 0;
  
}

/********************************************************************
   This function builds the Level structure.  This means that all
   the entities contained by 'ptr' that all the entities that were
   assigned a level of 0 will be placed in the LevelList[0] array.
   All the entities that were assigned a level of 1 will be placed 
   int the LevelList[1] array and so on.
**********************************************************************/

static void buildLevelStructure ( struct LevelType LevelList[], struct item *ptr )
{
    int level;
    int i;
    int tmp;
    struct item *p;
    
    
    for ( i=0; i < ptr -> numOfItemsContained; i++ )
    {
       p = indexList ( ptr -> ptrToConItems, i );
       level = p -> level - 1;
       tmp = LevelList [level].numOfItemsInLevel++;
       LevelList [level].list[tmp] = p;
    }
}


/********************************************************************
   Prints LevelStructure to debug file.
   Used for debugging purposes.
**********************************************************************/

static void printLevelStructure ( struct LevelType LevelList[], int numOfLevels )
{
   fprintf (debugFileNum, "\n\nPrinting the level structure\n" );
   fprintf (debugFileNum, "Displays the entities that were assigned to each level.\n\n" );
   
   {
      int i;
      int j;
      
      for ( i=0; i < numOfLevels; i++ )
      {
         if ( LevelList[i].numOfItemsInLevel != 0 )
         {
            fprintf ( debugFileNum, "Level : %d \n", i );
            for ( j=0; j <  LevelList[i].numOfItemsInLevel; j++ )
            {
               fprintf ( debugFileNum, "%30s %5d\n", LevelList[i].list[j] -> name, LevelList[i].list[j] -> indxIntoParConItemsArray );
            }
         } /* end if */
      } /* end for */
   } /* end scope */
}


/********************************************************************
   Scans through all the levels in the LevelList array and return
   the maximum number of items in a level.
**********************************************************************/

static int maxNumItemsInALevel ( struct LevelType LevelList[] , int numOfLevels )
{
   int i;
   int maxItems = 0;
   
   for ( i=0; i < numOfLevels; i++ )
   {
      if ( LevelList[i].numOfItemsInLevel > maxItems )
         maxItems = LevelList[i].numOfItemsInLevel;
   }
   return maxItems;
}


/********************************************************************
   Copies contents of matrix 'source' to matrix 'dest'.  
**********************************************************************/

static void copyMatrix ( int **dest, int **source, int dimension )
{
    int i;
    int j;
    for ( i=0; i < dimension; i++ )
    {
       for ( j=0; j < dimension; j++ )
          *(dest + j + i * dimension) = *(source + j + i * dimension);
    }      
}


/********************************************************************
  This function is a driver function that lays out the entities 
  contained by 'ptr'. The layout model used is the 'nested layout'.
**********************************************************************/

static void layoutContainedEntities ( ptr )
   struct item *ptr;
{
   
   
   int numOfLevels = 1;
   int numOfItemsCon = ptr -> numOfItemsContained;

   /** will contain relationships of graph with in container **/
   int adjacencyMatrix [numOfItemsCon][numOfItemsCon];
     
   /** will contain the relationships of the acyclic version of the graph **/
   int aMacyclic [numOfItemsCon][numOfItemsCon];
   
   buildAdjacencyMatrix ( (int **) adjacencyMatrix, ptr );
   
   calcUniqueRelationsTo ( (int **) adjacencyMatrix, ptr );
   sortEntitiesByUniqueRelationsTo ( ptr );

   /***
   adjacencyMatrix is built again because now the order of the entities has
   changed therefore you will get a different adjacencyMatrix ( the number
   of 1's in matrix is maintained just the positions might be different )
   ***/
     
   buildAdjacencyMatrix ( (int **) adjacencyMatrix, ptr );
  
   if ( debugMode ) 
      printMatrix ( (int **) adjacencyMatrix, ptr );
      
   /***
   copy adjacencyMatrix into aMacyclic because edges will be removed from the
   to make it acyclic.  The removing of edges will be done on aMacyclic.
   ***/
    
   copyMatrix ( (int **) aMacyclic, (int **) adjacencyMatrix, numOfItemsCon );    

   /*
   This will determine the levels of the entities contained by 'rootItemPtr'.
   It will also conver the graph represented by the matrix 'rMacyclic'
   from a cyclic graph to a acyclic graph if necessary.
   */
      
   determineLevelsInContainer ( ptr, (int **) aMacyclic, &numOfLevels );
   
   if ( debugMode )
   {
      fprintf (debugFileNum, "\n\n The following is the adjacency matrix after cycles have been removed.");
      printMatrix ( (int **) aMacyclic, ptr );
      printLevels ( ptr );
   }

   {
      struct LevelType LevelList[numOfLevels];
         
      initLevelList ( LevelList, numOfLevels );
         
      /** builds the LevelList data structure **/
      buildLevelStructure ( LevelList, ptr );
       
      if ( debugMode )
         printLevelStructure ( LevelList, numOfLevels );
         
      Control_SugiyamaLayout ( LevelList, (int **) adjacencyMatrix, numOfItemsCon, numOfLevels );

      Control_LevelLayout ( LevelList, (int **) adjacencyMatrix, numOfItemsCon, numOfLevels, maxNumItemsInALevel (LevelList,numOfLevels) );
 
      Control_PosDet ( ptr, ptr, LevelList, numOfLevels, maxNumItemsInALevel(LevelList, numOfLevels), 0, 0 );
   } /* end scope */
   
}


/********************************************************************
   This function visits every item to see if it contains other
   items/entities.  If a item has containment and 'item' not= 'exceptPtr'
   then the function 'layoutContainedEnties' is called to layout the
   contained entities.  

   Once exection gets to this point all the layouts will use the 
   'nested layout' model.
**********************************************************************/

static void visitEveryItemThatContainsStuff ( ptr, exceptPtr )
   struct item *ptr;
   struct item *exceptPtr;
{
   struct ListOfItems *p = ptr -> ptrToConItems;
   
   /*
   no matter what kind of layout model was used previously if program
   gets to here then the rest of the layouts will use nested layout
   model
   */
   layoutKind = nested;

   while (p)
   {
      if ( p -> itemPtr != exceptPtr )
      {
         if ( p -> itemPtr -> ptrToConItems )
         {
            layoutContainedEntities ( p -> itemPtr );
            visitEveryItemThatContainsStuff ( p -> itemPtr, exceptPtr );
         }
      }
      p = p -> next;
   }
}

/*********************************************************
  external function defenitions
**********************************************************/


/********************************************************************
   This is the driver function that lays out the entities in the top
   containment level only.
**********************************************************************/

void Control_DetermineLayout ( struct SystemPointers *graphPtrs )
{
   int numOfLevels = 1;

   if ( layoutKind == nested )
   {
     
      struct item *rootItemPtr = graphPtrs -> rootItemPtr;
      int numOfItemsCon = rootItemPtr -> numOfItemsContained;
      
      /** will contain relationships of graph with in container **/
      int adjacencyMatrix [numOfItemsCon][numOfItemsCon];
     
      /** will contain the relationships of the acyclic version of the graph **/
      int aMacyclic [numOfItemsCon][numOfItemsCon];
   
      assert ( rootItemPtr != NULL );
   
      /** first layout entities that $ROOT contains **/
      
      buildAdjacencyMatrix ( (int **) adjacencyMatrix, rootItemPtr );
   
      calcUniqueRelationsTo ( (int **) adjacencyMatrix, rootItemPtr );
      sortEntitiesByUniqueRelationsTo ( rootItemPtr );
   
   
      /***
      adjacencyMatrix is built again because now the order of the entities has
      changed therefore you will get a different adjacencyMatrix ( the number
      of 1's in matrix is maintained just the positions might be different )
      ***/
      
      buildAdjacencyMatrix ( (int **) adjacencyMatrix, rootItemPtr );
   
   
      if ( debugMode ) 
         printMatrix ( (int **) adjacencyMatrix, rootItemPtr );
      
      /***
      copy adjacencyMatrix into aMacyclic because edges will be removed from the
      to make it acyclic.  The removing of edges will be done on aMacyclic.
      ***/
      
      copyMatrix ( (int **) aMacyclic, (int **) adjacencyMatrix, numOfItemsCon );    
   
      /*
      This will determine the levels of the entities contained by 'rootItemPtr'.
      It will also conver the graph represented by the matrix 'rMacyclic'
      from a cyclic graph to a acyclic graph if necessary.
      */
      
      determineLevelsInContainer ( rootItemPtr, (int **) aMacyclic, &numOfLevels );
   
      if ( debugMode )
      {
         fprintf (debugFileNum, "\n\n The following is the adjacency matrix after cycles have been removed.");
         printMatrix ( (int **) aMacyclic, rootItemPtr );
         printLevels ( rootItemPtr );
      }
   
      {
         struct LevelType LevelList[numOfLevels];
         
         initLevelList ( LevelList, numOfLevels );
         
         /** builds the LevelList data structure **/
         buildLevelStructure ( LevelList, rootItemPtr );
       
         if ( debugMode )
            printLevelStructure ( LevelList, numOfLevels );
         
         Control_SugiyamaLayout ( LevelList, (int **) adjacencyMatrix, numOfItemsCon, numOfLevels );

         Control_LevelLayout ( LevelList, (int **) adjacencyMatrix, numOfItemsCon, numOfLevels, maxNumItemsInALevel (LevelList,numOfLevels) );
 
         Control_PosDet ( rootItemPtr, rootItemPtr, LevelList, numOfLevels, maxNumItemsInALevel(LevelList, numOfLevels), 0, 0 );
      } /* end scope */

      visitEveryItemThatContainsStuff ( rootItemPtr, NULL );
   }
   else if ( layoutKind == cliServ )
   {
        
      int numOfClients = 0;
      int numOfServers = 0;
      struct item *rootItemPtr = graphPtrs -> rootItemPtr;
      struct item *subSystemPtr = graphPtrs -> subSystemPtr;
      int numOfItemsCon = subSystemPtr -> numOfItemsContained;

      /** will contain relationships of graph with in container **/
      int adjacencyMatrix[numOfItemsCon][numOfItemsCon];
      
      /** will contain the relationships of the acyclic version of the graph **/
      int aMacyclic[numOfItemsCon][numOfItemsCon];
      
      assert ( subSystemPtr != NULL );
      assert ( rootItemPtr != NULL );
      
      buildAdjacencyMatrix ( (int **) adjacencyMatrix, subSystemPtr );

      calcUniqueRelationsTo ( (int **) adjacencyMatrix, subSystemPtr );
      sortEntitiesByUniqueRelationsTo ( subSystemPtr );
      
      /***
      adjacencyMatrix is built again because now the order of the entities has
      changed therefore you will get a different adjacencyMatrix ( the number
      of 1's in matrix is maintained just the positions might be different )
      ***/
      
      buildAdjacencyMatrix ( (int **) adjacencyMatrix, subSystemPtr );
      
      if ( debugMode )
         printMatrix ( (int **) adjacencyMatrix, subSystemPtr );
      
      /***
      copy adjacencyMatrix into aMacyclic because edges will be removed from the
      to make it acyclic.  The removing of edges will be done on aMacyclic.
      ***/
      
      copyMatrix ( (int **) aMacyclic, (int **) adjacencyMatrix, numOfItemsCon );    
   
      /*
      This will determine the levels of the entities contained by 'subSystemPtr'.
      It will also conver the graph represented by the matrix 'rMacyclic'
      from a cyclic graph to a acyclic graph if necessary.
      */
      
      determineLevelsInContainer ( subSystemPtr, (int **) aMacyclic, &numOfLevels );
      
      if ( debugMode )
      {
         fprintf (debugFileNum, "\n\n The following is the adjacency matrix after cycles have been removed.");
         printMatrix ( (int **) aMacyclic, subSystemPtr );
         printLevels ( subSystemPtr );
      }
      {
         struct LevelType LevelList[numOfLevels];
         
         initLevelList ( LevelList, numOfLevels );
         
         /** builds the LevelList data structure **/
         buildLevelStructure (LevelList,subSystemPtr);
      
         if ( debugMode )
            printLevelStructure ( LevelList, numOfLevels );
         
         Control_SugiyamaLayout ( LevelList, (int **) adjacencyMatrix, numOfItemsCon, numOfLevels );

         Control_LevelLayout ( LevelList, (int **) adjacencyMatrix, numOfItemsCon, numOfLevels, maxNumItemsInALevel (LevelList,numOfLevels) );
 
         Control_CliServLayout ( rootItemPtr, subSystemPtr, &numOfClients, &numOfServers );

         Control_PosDet ( rootItemPtr, subSystemPtr, LevelList, numOfLevels, maxNumItemsInALevel(LevelList, numOfLevels), numOfClients, numOfServers);
         
      } /* end scope */
      
      /*
      then see if any thing in a lower containment level needs to
      be layed out and do a layout
      */

      visitEveryItemThatContainsStuff ( subSystemPtr, NULL );
      visitEveryItemThatContainsStuff ( rootItemPtr, subSystemPtr );

   } /* end if */
}


   
