/*********************************************************
  This file is responsible for assigning each entity in 
  the graph attributes ( x,y,height,width ).  
**********************************************************/


#include <stdio.h>

#include "./../common/common.h"
#include "./../common/globalVars.h"
#include "PosDet.h"

/*********************************************************
  static variable defenitions
**********************************************************/

static float systemHeight;
static float systemWidth;

static float xInset;
static float yInset;

static float xBlock;
static float yBlock;

static float clientSlotHeight = 100;
static float serverSlotHeight = 100;

/***********************************  
  Declarations of static function 
************************************/
static void cliServInitialize();
static void nestedInitialize( struct item * );
static void determineWidthAndHeightOfConItems ( struct item *, int, int );
static void determineXandYofConItems ( struct LevelType *, int );
static void positionClients ( struct item * , int );
static void positionServers ( struct item * , int );
static void setAttributesOfSS ( struct item * );

/*********************************************************
  static function defenitions
**********************************************************/

static void cliServInitialize()
{
   systemHeight = WINDOW_HEIGHT - clientSlotHeight - serverSlotHeight;
   systemWidth = WINDOW_WIDTH - 50;
}

static void nestedInitialize( parentPtr )
   struct item *parentPtr;
{
   systemHeight = parentPtr -> height;
   systemWidth = parentPtr -> width;
}


static void determineWidthAndHeightOfConItems ( ptr, numberOfLevels, maxItemsInLevel )
   struct item *ptr;
   int numberOfLevels;
   int maxItemsInLevel;
{
   float boxHeight;
   float boxWidth;
   
   float maxBoxWidth;
   float maxBoxHeight;
   
   xBlock = systemWidth / ( maxItemsInLevel + 1);
   yBlock = systemHeight / ( numberOfLevels + 1 );
   
   maxBoxWidth = xBlock * WIDTH_SCALE;
   maxBoxHeight = yBlock * HEIGHT_SCALE;

   boxWidth = maxBoxWidth;
   boxHeight = boxWidth * HEIGHT_TO_WIDTH_RATIO;
      
   if ( boxHeight > maxBoxHeight )
   {
      boxHeight =  maxBoxHeight;
      boxWidth = boxHeight * WIDTH_TO_HEIGHT_RATIO;
   }   

   xInset = xBlock - boxWidth;
   yInset = yBlock - boxHeight;
   
   {
      struct ListOfItems *lp = ptr -> ptrToConItems;
      while ( lp )
      {
         lp -> itemPtr -> height = boxHeight;
         lp -> itemPtr -> width = boxWidth;
         lp = lp -> next;
      }
   }
   
}

static void determineXandYofConItems ( LevelList, numberOfLevels )
   struct LevelType LevelList[];
   int numberOfLevels;
{
   float curXpos = 0.0;
   float curYpos = 0.0;
   int i,j;

   for ( i=0; i < numberOfLevels; i++ )
   {
      for ( j=0; j < LevelList[i].numOfItemsInLevel; j++ )
      {
         if ( numberOfLevels == 1 )
            curXpos = ( LevelList[i].list[j] -> desiredPosition - 1 ) * xBlock;
         else
            curXpos = ( LevelList[i].list[j] -> actualPosition - 1 ) * xBlock;
            
         LevelList[i].list[j] -> x = curXpos + xInset;
         LevelList[i].list[j] -> y = curYpos + yInset;
      }
      curYpos += yBlock;
   }
}


static void positionClients ( rootItemPtr, numOfClients )
   struct item *rootItemPtr;
   int numOfClients;
{
   if ( numOfClients > 0 )
   {
      float clientSlotWidth = WINDOW_WIDTH / numOfClients;
      float xInset = clientSlotWidth * 0.5 / 2;
      float yInset = clientSlotHeight * 0.5 / 2;
      struct ListOfItems *ptr = rootItemPtr -> ptrToConItems;
      struct item *p;
      
      while ( ptr )
      {  
         p = ptr -> itemPtr;
         
         if ( p -> kindOfItemInModel == client ) 
         {
            p -> width = clientSlotWidth - ( xInset * 2 );
            p -> height = p -> width * HEIGHT_TO_WIDTH_RATIO;
            if ( p -> height > MAX_CLIENT_OR_SERVER_HEIGHT )
            {
               p -> height = MAX_CLIENT_OR_SERVER_HEIGHT;
               p -> width = p -> height * WIDTH_TO_HEIGHT_RATIO;
            }
            p -> x = ( p -> actualPosition - 1 ) * clientSlotWidth + xInset;
            p -> y = yInset;
         }
         ptr = ptr -> next;
      }
   }
}


static void positionServers ( rootItemPtr, numOfServers )
   struct item *rootItemPtr;
   int numOfServers;
{
   if ( numOfServers > 0 )
   {
      float serverSlotWidth = WINDOW_WIDTH / numOfServers;
      float xInset = serverSlotWidth * 0.5 / 2;
      float yInset = serverSlotHeight * 0.5 / 2;
      struct ListOfItems *ptr = rootItemPtr -> ptrToConItems;
      struct item *p;
      
      while ( ptr )
      {  
         p = ptr -> itemPtr;
         
         if ( p -> kindOfItemInModel == server ) 
         {
            p -> width = serverSlotWidth - ( xInset * 2 );
            p -> height = p -> width * HEIGHT_TO_WIDTH_RATIO;
            if ( p -> height > MAX_CLIENT_OR_SERVER_HEIGHT )
            {
               p -> height = MAX_CLIENT_OR_SERVER_HEIGHT;
               p -> width = p -> height * WIDTH_TO_HEIGHT_RATIO;
            }
            p -> x = ( p -> actualPosition - 1 ) * serverSlotWidth + xInset;
            p -> y = WINDOW_HEIGHT - yInset - p -> height;
         }
         ptr = ptr -> next;
      }
   }
}

static void setAttributesOfSS ( subSystemPtr )
   struct item *subSystemPtr;
{
   subSystemPtr -> x = ( WINDOW_WIDTH - systemWidth ) / 2;
   subSystemPtr -> y = clientSlotHeight;
   subSystemPtr -> height = systemHeight;
   subSystemPtr -> width = systemWidth;
}


void Control_PosDet (parentPtr, subSystemPtr, LevelList, numberOfLevels, maxItemsInLevel, numOfClients, numOfServers )
   struct item *parentPtr;
   struct item *subSystemPtr;
   struct LevelType LevelList[];
   int numberOfLevels;
   int maxItemsInLevel;
   int numOfClients;
   int numOfServers;
{
   if ( layoutKind == cliServ )
   {
      cliServInitialize();
      positionClients ( parentPtr, numOfClients );
      positionServers ( parentPtr, numOfServers );
      setAttributesOfSS ( subSystemPtr );
      determineWidthAndHeightOfConItems ( subSystemPtr, numberOfLevels, maxItemsInLevel );
      determineXandYofConItems ( LevelList, numberOfLevels );
   }
   else if ( layoutKind == nested )
   {
      nestedInitialize( parentPtr );
      determineWidthAndHeightOfConItems ( parentPtr, numberOfLevels, maxItemsInLevel );
      determineXandYofConItems ( LevelList, numberOfLevels );
   }
}
   
